/* RCS Information:
 *
 * $Id$
 * $Revision$
 * $Date$
 * $Author$
 * $Locker$
 * 
 */

/******************************************************************************
 *
 *  Copyright (C) 2001, Bill Pierce
 *  This file is part of the timeout_action library.
 *
 *  The timeout_action library is free software; you can 
 *  redistribute it and/or modify it under the terms of the 
 *  GNU Library General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  The timeout_action library is distributed in the hope that 
 *  it will be useful, but WITHOUT ANY WARRANTY; without even 
 *  the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 *  PARTICULAR PURPOSE.  See the GNU Library General Public 
 *  License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with the timeout action libary; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Bill Pierce can be contacted at:
 *  7862 NE 148th St.
 *  Kenmore, WA 98028
 *
 * Name --  timeout_action.c
 *
 * Purpose --  Implementation of portable functions defined in 
 *             timeout_action.h on Windows and UNIX platforms. 
 *             This function allows a function to be executed
 *             such that the calling thread blocks until either
 *             the function returns or a specified amount of time
 *             elapses.
 *
 * Remarks --  
 *
 *
 * Known bugs --  
 *
 * Applicable Invariants -- 
 *
 * Supported Platforms  -- 
 *
 * Win32 circa NT, 2000,    compile under cl,  define MS_WIN32 
 * GNU/Linux circa RH 6.2,  compile under gcc, define __UNIX__, __USE_POSIX__
 * Solaris circa 2.7,       compile under cc   define __UNIX__, __USE_POSIX__
 *
 * Language Version --  C - GNU/Linux 
 *
 * Author:  Bill Pierce
 * 
 *****************************************************************************/

 #include <stdio.h>
 #include "timeout_action.h"

 #ifdef __UNIX__
   #include <unistd.h>
   #include <stdlib.h>   /* provides malloc(), free() */
   #include <time.h>     /* provides nanosleep() */
   #include <errno.h>    /* provides errno values. */
   #include <string.h>   /* provides strerror() */
 #endif

 #ifdef __USE_POSIX__
   #include <semaphore.h>
   #include <pthread.h>
 #endif

 #ifdef MS_WIN32
   #include <windows.h>
 #endif

/* The following provide implementations on UNIX/POSIX platforms. */
#ifdef __UNIX__
  #ifdef __USE_POSIX__

 /*
 ** Function and structure prototypes.
 */

 /*
  * This is a structure used in the pthreads implementation of 
  * timeout_action.  It is passed between threads of control. 
  * It contains the parameters and return values that are passed 
  * to the function to be executed as well as the timeout value
  * that is passed to the timer function and semaphore and mutex
  * handles that are used to synchronize the threads. 
  */
 typedef struct timeout_action_control {
    void* (*p_action) (void*);        /* A function pointer. */
    void* arg;                        /* Argument to the function. */
    void* return_value;               /* Return value of the function. */
    unsigned int timeout_in_ms;       /* Timeout value for the function. */
    sem_t timeout_semaphore;          /* Semaphore to signal when done. */
    pthread_mutex_t control_mutex;    /* Mutex used to coordinate threads. */
    volatile pthread_t signalling_thread_id;   /* id of the winning thread. */
    volatile int signalling_error_code; /* num errors in thread func. */
 } TIMEOUT_ACTION_CONTROL;

 /*
 ** Global variables 
 */

 /*
 ** Function implementations. 
 */

 /* Dynamically allocate and initialize a TIMEOUT_ACTION_CONTROL 
  * structure. This includes calling initialization routines for
  * the contained semaphore and mutex. 
  *
  * Return a pointer to this structure or NULL if the
  * operation fails.
  */
 TIMEOUT_ACTION_CONTROL *create_action_control(void* p_function (void*),
                                               void* p_arg,
                                               unsigned int timeout_in_ms
                                              ) {
  
    TIMEOUT_ACTION_CONTROL * p_action_control;

    p_action_control = (TIMEOUT_ACTION_CONTROL*) 
                       malloc(sizeof(TIMEOUT_ACTION_CONTROL));

    if (NULL != p_action_control) {
       #ifdef _DEBUG
          printf("Alloced control at %p\n", p_action_control);
       #endif
       /* Initialize the structure. */
       p_action_control->p_action             = p_function;
       p_action_control->arg                  = p_arg;
       p_action_control->return_value         = NULL;
       p_action_control->timeout_in_ms        = timeout_in_ms;
       p_action_control->signalling_thread_id = 0;
       p_action_control->signalling_error_code= 0;
       sem_init(&p_action_control->timeout_semaphore, 0, 0);
       pthread_mutex_init(&p_action_control->control_mutex, NULL);
    } 
    return p_action_control;
 } /* End of create_action_control. */

 /* Uninitialize contained mutex and semaphores free the 
  * memory structure pointed to by p_action_control.
  * Return 1 on success, 0 on failure.
  */
 int destroy_action_control(TIMEOUT_ACTION_CONTROL* p_action_control) {

    int cumulative_error_count = 0;

    /* Clean up the semaphore. */
    if (0 != sem_destroy(&p_action_control->timeout_semaphore)) {
       #ifdef _DEBUG
          fprintf(stderr, 
                  "destroy_action_control: Error destroying semaphore\n"
                 );
       #endif
       cumulative_error_count++;
    }
    /* Clean up the mutex. */
    if (0 != pthread_mutex_destroy(&p_action_control->control_mutex)) {
       #ifdef _DEBUG
          fprintf(stderr, 
                  "destroy_action_control: Error destroying mutex\n"
                 );
       #endif
      cumulative_error_count++;
    }

    #ifdef _DEBUG
       printf("About to free control at %p\n", p_action_control);
    #endif

    free(p_action_control);

    if (0 != cumulative_error_count) {
       return 0; 
    } else {
       return 1;
    }
 } /* End of destroy_action_control. */

 /*
  * This is an internal function used by timeout_action. 
  * This is a function that can be executed in a separate
  * thread.  Its signature is that defined by the 
  * pthread_create function.  It attempts to sleep for the 
  * duration specified in timeout_in_ms. Once the sleep
  * completes, it attempts to attain a lock to a mutex and
  * send a post signal to the semaphore designated by 
  * timeout_semaphore. 
  *
  * arg should contain a pointer to a TIMEOUT_ACTION_CONTROL
  * structure.
  *
  * This function returns a NULL pointer.  If internal errors
  * were generated during this call it sets the error code
  * in p_timeout_control->signalling_error_code to a nonzero value. 
  */
 void* timer_thread_function(void* arg) {

    TIMEOUT_ACTION_CONTROL *p_timeout_control;
    
    unsigned int    total_ms_to_sleep;
    time_t          seconds_to_sleep;
    long            nanoseconds_to_sleep;
    struct timespec tv;
    int cumulative_error_count = 0;
    int destroy_after_unlock   = 0;

    #ifdef _DEBUG
    int system_error_code      = 0;
    #endif

    p_timeout_control = (TIMEOUT_ACTION_CONTROL*) arg;

    /* Put this thread in a detached state so it will 
     * relesae its resources when it exits.
     */
    if (pthread_detach(pthread_self())) {
       #ifdef _DEBUG
          fprintf(stderr, "timer_thread_function: Error detaching thread\n");
       #endif
       cumulative_error_count++;
    }

    /* Compute the sleep structure for nanosleep. */
    total_ms_to_sleep = p_timeout_control->timeout_in_ms;
    seconds_to_sleep = (time_t) (total_ms_to_sleep/1000); 
    nanoseconds_to_sleep = 1000000 * (total_ms_to_sleep - 
                                   1000 * seconds_to_sleep);
    tv.tv_sec  = seconds_to_sleep;
    tv.tv_nsec = nanoseconds_to_sleep; 

    /* Go to sleep for the specified wait period. */
    if (0 != nanosleep(&tv, &tv)) {
       #ifdef _DEBUG
          fprintf(stderr, "timer_thread_function: nanosleep failed\n");
          system_error_code = errno;
          fprintf(stderr, "System error code was %s\n",
                  strerror(system_error_code));
       #endif 
       cumulative_error_count++;
    }

    #ifdef _DEBUG
       printf("Done sleeping.\n");
    #endif

    /* We've finished the race. Lock the control mutex. */
    if(pthread_mutex_lock(&p_timeout_control->control_mutex)) {
       cumulative_error_count++;
    } else {
       #ifdef _DEBUG
        printf("Timer thread obtained mutex.\n");
       #endif
       /* If we're first, Signal the calling thread that we're done. */
       if (0 == p_timeout_control->signalling_thread_id) { 
          p_timeout_control->signalling_thread_id = pthread_self();
          p_timeout_control->signalling_error_code = cumulative_error_count;
          sem_post(&p_timeout_control->timeout_semaphore);
       } else {
          destroy_after_unlock = 1;
       }
       #ifdef _DEBUG
        printf("Timer thread about to release mutex.\n");
       #endif
       /* Unlock the control mutex. Hope this does not fail.*/
       pthread_mutex_unlock(&p_timeout_control->control_mutex);
       #ifdef _DEBUG
          printf("Timer thread released mutex.\n");
       #endif
    }

    /* We were the last one out. We'll turn off the lights. */
    if (destroy_after_unlock) {
       /* This sleep is intended to give the main 
        * thread time to pick up the result from the 
        * control structure before we destroy it,
        * but there is still, technically a race here.
        */
       sleep(1);
       destroy_action_control(p_timeout_control);
    }

    pthread_exit(NULL);

    return NULL;
 } /* End of timer_thread_function. */ 

 /*
  * This is an internal function used by timeout_action. 
  * This is a function that can be executed in a separate
  * thread.  Its signature is that defined by the 
  * pthread_create function.  It attempts to execute 
  * the action pointed to by p_timeout_control->p_action.  Once this action
  * returns, it attempts to attain a lock to a mutex and
  * send a post signal to the semaphore designated by 
  * p_action_control->timeout_semaphore. 
  *
  * arg should contain a pointer to a TIMEOUT_ACTION_CONTROL structure.
  *
  * This function returns a NULL pointer.  If internal errors
  * were generated during this call it sets the error code
  * in p_action_control->signalling_error_code to a nonzero value. 
  */
 void* signalling_action(void* arg) {

    TIMEOUT_ACTION_CONTROL *p_timeout_control;
    int cumulative_error_count = 0;
    int destroy_after_unlock   = 0;

    p_timeout_control = (TIMEOUT_ACTION_CONTROL*) arg;

    /* Put this thread in a detached state so it will 
     * release its resources when it exits.
     */
    if (pthread_detach(pthread_self())) {
       #ifdef _DEBUG
          fprintf(stderr, "signalling_action: Error detaching thread\n");
       #endif
       cumulative_error_count++;
    }

    /* Execute the function at the function pointer. */
    #ifdef _DEBUG
      printf("Starting action.\n");
    #endif
    p_timeout_control->return_value =
    p_timeout_control->p_action(p_timeout_control->arg); 
    #ifdef _DEBUG
      printf("Done with action.\n");
    #endif

    /* We've finished the race. Lock the control mutex.*/
    if(pthread_mutex_lock(&p_timeout_control->control_mutex)) {
       cumulative_error_count++;
    } else {
       #ifdef _DEBUG
          printf("Action thread obtained mutex.\n");
       #endif
       /* If we're first, signal the calling thread that we're done. */
       if (0 == p_timeout_control->signalling_thread_id) { 
          p_timeout_control->signalling_thread_id  = pthread_self();
          p_timeout_control->signalling_error_code = cumulative_error_count;
          sem_post(&p_timeout_control->timeout_semaphore);
       } else {
          /* Otherwise, set a flag so we can clean up. */
          destroy_after_unlock = 1;
       }
       #ifdef _DEBUG
          printf("Action thread about to release mutex.\n");
       #endif
       /* Unlock the mutex. Hope this does not fail.*/
       pthread_mutex_unlock(&p_timeout_control->control_mutex);
    }

    #ifdef _DEBUG
       printf("Action thread let go of mutex.\n");
    #endif

    /* We were the last one out, we'll turn off the lights.*/
    if (destroy_after_unlock) {
       /* This sleep is intended to give the main 
        * thread time to pick up the result from the 
        * control structure before we destroy it,
        * but there is still, technically a race here.
        */
       sleep(1);
       destroy_action_control(p_timeout_control);
    }

    pthread_exit(NULL);

    return NULL;
 } /* End of signalling_action */

  /* UNIX pthreads implementation of the function defined in
   * timeout_action.h.  It works by attempting to start two
   * threads; one to sleep for the specified timeout interval
   * and one to concurrently execute the function pointed to 
   * by p_function.  After starting these threads, the 
   * calling thread blocks until it is sent a semaphore by 
   * the winning thread.  The losing thread finishes silently
   * and cleans up the control structure when it is done 
   * executing. 
   */
 int timeout_action( void* p_function (void*), 
                     void* p_arg,
                     unsigned int timeout_in_ms,
                     void** pp_return_value 
                   ) {

    int return_code = TIMEOUT_ACTION_CODE_ERROR;
    int status_code = 0;

    #ifdef _DEBUG
    int system_error_code = 0;
    #endif

    /* This is used to keep track of any unrecoverable errors 
     * that might occur. If there was an error, at least we'll
     * know it.
     */
    int cumulative_error_count = 0;

    /* The thread ids. */
    pthread_t action_thread_id;
    pthread_t timer_thread_id;

    /* A control structure passed as a parameter between threads. */
    TIMEOUT_ACTION_CONTROL* p_timeout_control; 
    
    /* Initialize the control structure. Unless an error occurs, this
     * structure will be destroyed by the losing thread once it completes.
     */
    p_timeout_control = create_action_control(p_function, 
                                              p_arg, 
                                              timeout_in_ms);

    /* Check for failure to initialize. */
    if (NULL == p_timeout_control) {
       #ifdef _DEBUG
	 fprintf(stderr, 
                 "timeout_action: Failed to initialize control structure.\n");
       #endif
       return TIMEOUT_ACTION_CODE_ERROR;
    }

    /* Start the timer thread. */
    #ifdef _DEBUG
       printf("starting the timer thread\n");
    #endif

    status_code = pthread_create(&timer_thread_id, 
                                 NULL, 
                                 &timer_thread_function, 
                                 p_timeout_control);
    if (0 != status_code) { 
       #ifdef _DEBUG 
          fprintf(stderr, "timeout_action: Error starting timer thread.\n");
          system_error_code = errno;
          fprintf(stderr, "System error code was %s\n",
                  strerror(system_error_code));
       #endif
       destroy_action_control(p_timeout_control);
       return TIMEOUT_ACTION_CODE_ERROR; 
    }

    /* Start the action. */
    #ifdef DEBUG
       printf("starting the action thread");
    #endif

    status_code = pthread_create(&action_thread_id, 
                                 NULL, 
                                 &signalling_action, 
                                 p_timeout_control);

    if (0 != status_code) { 
       #ifdef _DEBUG 
          fprintf(stderr, "timeout_action: Error starting action thread\n");
          system_error_code = errno;
          fprintf(stderr, "System error code was %s\n",
                  strerror(system_error_code));
       #endif
       /* Mark the control structure so the timer thread thinks 
        * someone else won the race.
        */
       if (pthread_mutex_lock(&p_timeout_control->control_mutex)) {
         p_timeout_control->signalling_thread_id = pthread_self();
         pthread_mutex_unlock(&p_timeout_control->control_mutex);
       }
       destroy_action_control(p_timeout_control);
       return TIMEOUT_ACTION_CODE_ERROR; 
    }

    /* Wait for either the timer or the action to complete.
     */
    #ifdef _DEBUG
       printf("Main waits for semaphore signal\n");
    #endif

    sem_wait(&p_timeout_control->timeout_semaphore);

    #ifdef _DEBUG
      printf("Main is done waiting on the semaphore.\n");
    #endif

    /* See which thread won the race. */
    if (pthread_equal(action_thread_id, 
                      p_timeout_control->signalling_thread_id
                     )
       ) {
       #ifdef _DEBUG
          printf("Action won\n");
       #endif
       /* Pick up the return value. There is a potential race 
        * here if the losing thread cleans up the control 
        * structure before we can get the results out.
        */
       *(pp_return_value) = p_timeout_control->return_value;
       return_code  = TIMEOUT_ACTION_CODE_SUCCESS;
       #ifdef _DEBUG
          printf("Done processing action won.\n");
       #endif
    } else {
       #ifdef _DEBUG
          printf("timer won\n");
       #endif
       return_code  = TIMEOUT_ACTION_CODE_TIMEOUT;
       #ifdef _DEBUG
          printf("Done processing timer won.\n");
       #endif
    } /* End if on pthread_equal. */

    /* See if there were errors in the thread that won. */ 
    cumulative_error_count += p_timeout_control->signalling_error_code; 

    /* If there was an error along the way, overwite any other return code. */
    if (cumulative_error_count) {
       return_code = TIMEOUT_ACTION_CODE_ERROR;
       #ifdef _DEBUG
       fprintf(stderr, 
               "Cumulatvie error count is %d\n", 
               cumulative_error_count); 
       #endif
    }

    return return_code;
 }
  #endif  /* __USE_POSIX__ */
#endif /* __UNIX__ */


/* The following section provides an implementation on WIN32. */
#ifdef MS_WIN32

 typedef struct timeout_action_control {
    void* (*p_action) (void*);        /* A function pointer. */
    void* arg;                        /* Argument to the function. */
    void* return_value;               /* Return value of the function. */
 } WRAPPED_ACTION_CONTROL;

 /*
  * This function wraps the action, handling return values, etc.
  */
 void wrapped_action(void* arg) {

    WRAPPED_ACTION_CONTROL *p_action_control;

    p_action_control = (WRAPPED_ACTION_CONTROL*) arg;

    #ifdef _DEBUG
      printf("Starting action.\n");
    #endif
    p_action_control->return_value =
    p_action_control->p_action(p_action_control->arg); 
    #ifdef _DEBUG
      printf("Done with action.\n");
    #endif
    return;
 }

/* 
 * MS Win32 form of the method defined in timeout_action.h  Uses Win32 
 * Threading calls.
 */
 int timeout_action( void* p_function (void*), 
                     void* p_arg,
                     unsigned int timeout_in_ms,
                     void** pp_return_value 
                   ) {

   int                    return_code;
   WRAPPED_ACTION_CONTROL action_control;
   HANDLE                 threadHandle;
   DWORD                  threadId;

   /* Initialize the structure. */
   action_control.p_action             = p_function;
   action_control.arg                  = p_arg;
   action_control.return_value         = NULL;

   threadHandle = CreateThread(0,
                               0,
                               (LPTHREAD_START_ROUTINE) wrapped_action,
                               (VOID *) &action_control, 
                               0, 
                               &threadId
                              );             
   if (0 != threadHandle) {
      DWORD status;

      #ifdef _DEBUG
         printf("Waiting for action to complete.\n");
      #endif
      // Wait for the thread to do its work or time-out.
      status = WaitForSingleObject(threadHandle, 
                                   timeout_in_ms 
                                  );

      // Release the thread handle.
      CloseHandle(threadHandle);

      // See if we timed out or executed normally.
      if ((WAIT_ABANDONED == status) ||
          (WAIT_TIMEOUT   == status)    ) {
         #ifdef _DEBUG
            printf("Action function timed out.\n");
         #endif
         return_code = TIMEOUT_ACTION_CODE_TIMEOUT;
      } else if (WAIT_OBJECT_0 == status) {
         #ifdef _DEBUG
            printf("Action function ran to completion.\n");
         #endif
         return_code = TIMEOUT_ACTION_CODE_SUCCESS;
         *(pp_return_value) = action_control.return_value;
      } else {
         #ifdef _DEBUG
           fprintf(stderr, 
           "timeout_action: wait failed.\n");
         #endif
         return_code = TIMEOUT_ACTION_CODE_ERROR;
      }

   } else {
      #ifdef _DEBUG
         fprintf(stderr, 
         "timeout_action: Unable to create thread to do timeout action.\n");
      #endif
      return_code = TIMEOUT_ACTION_CODE_ERROR;
   }
    
   return return_code;

} /* End timeout_action Win32. */

#endif  /* MS_WIN32 */

 /*****************************************************************************
 ** $Log: $
 *****************************************************************************/

